package Hadoop::IO;
$Hadoop::IO::VERSION = '0.003';
use 5.010;
use strict;
use warnings;

1;

__END__

=pod

=encoding UTF-8

=head1 NAME

Hadoop::IO

=head1 VERSION

version 0.003

=head1 SYNOPSIS

None

=head1 NAME

Hadoop::IO - This is not a real code and only purpose is to keep PAUSE happy

=head1 AUTHORS

=over 4

=item *

Philippe Bruhat

=item *

Sabbir Ahmed

=item *

Somesh Malviya

=item *

Vikentiy Fesunov

=back

=head1 COPYRIGHT AND LICENSE

This software is copyright (c) 2023 by Booking.com.

This is free software; you can redistribute it and/or modify it under
the same terms as the Perl 5 programming language system itself.

=cut
